package IkiWiki::Plugin::sar;
use warnings;
use strict;
use Carp;
use English q(-no_match_vars);
use List::MoreUtils qw(uniq);
use Data::Dumper;

use IkiWiki 2.00;

our $VERSION = '0.0.8';

#
#   out global dictionary: every key is the search text; first subkey is the
#   replace text on the first match and next subkey is the replace text on the
#   others matches 
my %_dictionary         =   ();

my %_local_dict         =   ();     # pages local dictionary
my %_page_counters      =   ();     
my %_preprocess_page    =   ();

my $_sar_mainpage       =   q(sar); # global page for dictionary
my $_sar_pagespec       =   q(*);   # enalbe on all pages by default

my $_rbound             =   q(--);  # right boundary
my $_lbound             =   q(--);  # left boundary

### INTERFACE SUB ###
# Usage         : IkiWiki::Plugin::sar::import()
# Purpose       : Declare the hooks to ikiwiki.
# Returns       : nothing
# Parameters    : nothing 
# Throws        : no exceptions
# Commments     : none
# See also      : n/a

sub import { #{{{
	hook(type => "checkconfig", id => "sar", call => \&checkconfig);
	hook(type => "preprocess",  id => "sar", call => \&preprocess);
	hook(type => "filter",      id => "sar", call => \&filter);
} # }}}
    
### INTERFACE SUB ###
# Usage         : IkiWiki::Plugin::sar::checkconfig();
# Purpose       : - Read the global parameters from ikiwiki config.
#               : - Load and preprocess the global dictionary page if exists.
# Returns       : nothing
# Parameters    : nothing
# Throws        : no exceptions
# Commments     : none
# See also      : n/a

sub checkconfig { #{{{
    # Read the global parameters
    if (defined $config{sar_pagespec}) {
        $_sar_pagespec = $config{sar_pagespec};
    }
    else {
        $_sar_pagespec = undef;
    }

    # Preprocess the global dictionary page if exists
    if (my $path_to_dict = srcfile("${_sar_mainpage}.mdwn")) {
        IkiWiki::preprocess($_sar_mainpage, $_sar_mainpage,
                readfile($path_to_dict));
    }
    else {
        debug(sprintf("sar main page '%s' not found", $_sar_mainpage));
    }
} #}}}

### INTERFACE SUB ###
# Usage         : IkiWiki::Plugin::sar::preprocess( %params );
# Purpose       : Add a sar definition to the dictionary.
# Returns       : Text to show 
# Parameters    : - search  => "Text to search "
#               : - first   => "Replace the first time"
#               : - next    => "Replace the others"
#               : - replace => "Replace all matches"
#               : - page    => "Page name"
# Throws        : no exceptions
# Commments     : If the page matches to the main page, the sar definition is
#               : added to the global dictionary. Otherwise the sar is added
#               : to a local dictionary for the page.
#               : If the sar definition is empty (i mean missing first and
#               : next parameters) and the page is not the main page, the
#               : definition is disabled for the page scope.
# See also      : n/a

sub preprocess (@) { #{{{
	my  %params         =   @_;
    my  $search_text    =   $params{search};

    # setting defaults 
    if (exists($params{'next'})) {
        $params{'replace'} = $params{'next'};
    }
    if (not exists($params{'first'})) {
        $params{'first'} = $params{'replace'};
    }

    # if we are in the main dictionary page ...
    if ($params{page} eq $_sar_mainpage) {
        if (defined $search_text) {
            # save in the dictionary
            _add_to_dict( $search_text, $params{'first'}, $params{'replace'});

            # return a explanation
            return _show_replace($search_text);
        }            
    }
    # No, we are in other page 
    else {
        # only if defined a search text ...
        if (defined $search_text) {
            _add_to_local( $params{page}, $search_text, $params{'first'},
                        $params{'replace'});
        }

        # return nothing            
        return;
    }        
} # }}}

### INTERNAL UTILITY ###
# Usage         : _add_to_dict( $search, $page, $first, $next );
# Purpose       : Add an entry to the dictionary
# Returns       : - Reference to the entry
# Parameters    : - Search text (the hash key)
#               : - Page name (if the entry is specific for a page)
#               : - Replace text the first match 
#               : - Replace text the next matches
# Throws        : no exceptions
# Commments     : none
# See also      : n/a

sub _add_to_dict {
    my  $search_text    =   shift;
    my  $first_match    =   shift;  # optional
    my  $next_match     =   shift;  # optional

    if (not exists $_dictionary{$search_text}) {
        $_dictionary{$search_text} = {
                'first'     =>  undef,
                'replace'   =>  undef,    
            };
    }

    # pointer to the entry (for readability)
    my $entry_dict = $_dictionary{$search_text};

    # save the replace texts
    $entry_dict->{'first'} = $first_match;
    $entry_dict->{'replace'} = $next_match;

    return $entry_dict;
}

sub _add_to_local {
    my  $page_name      =   shift;
    my  $search_term    =   shift;
    my  $first_match     =   shift;
    my  $next_match      =   shift;

    if (not exists $_local_dict{$page_name}) {
        # new entry for the page 
        $_local_dict{$page_name} = {
            'terms'     =>  {},         # special terms 
            };
    }
    if (not exists $_local_dict{$page_name}->{terms}->{$search_term}) {
        $_local_dict{$page_name}->{terms}->{$search_term} = {
                'first'     =>  undef,
                'replace'   =>  undef,
            };
    }
                
    my $entry = $_local_dict{$page_name}->{terms}->{$search_term};

    # save the new entry parameters
    $entry->{'first'}   =   $first_match;
    $entry->{'replace'} =   $next_match;

    return $_local_dict{$page_name};
}

sub _init_match_counters {
    %_page_counters = ();

    return;
}

sub _inc_counter {
    my  $term       =   shift;

    if (exists $_page_counters{$term}) {
        $_page_counters{$term}++;
    }
    else {
        # initialize the term counter 
        $_page_counters{$term} = 1;
    }

    return $_page_counters{$term};
}

### INTERNAL UTILITY ###
# Usage         : _show_replace( $search_text )
# Purpose       : Build a human description for the search text. Useful only
#               : in the main sar page.
# Returns       : Text with the description
# Parameters    : - Search text for lookup in the internal dictionary.
# Throws        : no exceptions
# Commments     : none
# See also      : n/a

sub _show_replace {
    my  $entry      =   shift;

    my $msg = sprintf("Search '%s%s%s' ", $_lbound, $entry, $_rbound);

    # if we have a text for the first appearance         
    if (defined $_dictionary{$entry}->{'first'}) {
        $msg .= sprintf("replaced by '%s'", $_dictionary{$entry}->{'first'});

        if (defined $_dictionary{$entry}->{'replace'}) {
            $msg .= sprintf(" and by '%s'",
                $_dictionary{$entry}->{'replace'});
        }
    }
    # no, we have a universal replace
    elsif (defined $_dictionary{$entry}->{'replace'}) {
        $msg .= sprintf("replaced by %s", $_dictionary{$entry}->{'replace'});
    }
    # error for this entry
    else {
        $msg .= sprintf('missing replace parameters');
    }        

    return $msg;
}

### INTERFACE SUB ###
# Usage         : IkiWiki::Plugin::sar::filter( %params )
# Purpose       : Search and replace texts in a ikiwiki source page.
# Returns       : Filtered content's page 
# Parameters    : - page        => "source page name"
#               : - destpage    => "optional destination page name"
#               : - content     => "page content text"
# Throws        : no exceptions
# Commments     : The function check first that the page name match with the global
#               : pagespec (if exists) and second lookup the name in the local
#               : dictionary.
# See also      : n/a

sub filter () { #{{{
    my %params      =   @_;
    my  $content    =   $params{content};

    # don't work on sar main page 
    return $content if $params{page} eq $_sar_mainpage;

    # if the page match a pagespec 
    if ((defined $_sar_pagespec 
         and pagespec_match( $params{page}, $_sar_pagespec )) ) {
        # if the page contains a sar directive ... 
        if (not $_preprocess_page{ $params{page} } and
           $content =~ m{[^\\]\[\[sar}xms) {

           # preprocess the content before filtering
           debug(sprintf("sar: force preprocess of %s page", $params{page}));
           IkiWiki::preprocess($params{page}, $params{destpage}, $content); 
           $_preprocess_page{ $params{page} } = 1;
        }

        debug("sar: filtering the page " . $params{page});

        # add a dependency on the sar main page
        add_depends($params{page}, $_sar_mainpage);

        # initialize the match counters for the page
        _init_match_counters();

        # work on the content page
        $content = _search_and_replace( $params{page}, $content); 

        # and fix the escaped sar expressions
        $content = _fix_escaped_expressions( $content );
    }

    return $content;
} #}}}

sub _search_and_replace {
    my  $page_name              =   shift;
    my  $page_content           =   shift;
    my  $depends_on_dictionary  =   0;

    while ($page_content =~ m{(\\?)--(\w+)--}xms) {
        $page_content =~ s{(\\?)--(\w+)--}{ 
                        _select_replace( $page_name, $1, $2)
                         }xmes;
    }

    return $page_content;
}

sub _fix_escaped_expressions {
    my  $content    =   shift;

    $content =~ s{--\s(\w+)\s--}{sprintf("--%s--",$1)}xmsge;

    return $content;
}

sub _msg_error {
    my  $page_name  =   shift;

    return sprintf("\n[[sar: undefined regular expression]]\n");
}

sub _select_replace {
    my  $page_name      =   shift;
    my  $escape         =   shift;
    my  $term           =   shift;
    my  $replace        =   undef;

    # is it escaped ? 
    if (not $escape) {
        # search a replace for the term
        $replace = _replace_term( $page_name, $term );
    }

    # if not we have a replace ...
    if (not $replace) {
        # return a literal text
        $replace = sprintf('-- %s --', $term);
    }

    return $replace;
}

sub _replace_term {
    my  $page_name      =   shift;
    my  $term           =   shift;

    # cleaning the term
    $term =~ s{--}{}xmsg;

    # first step: check the term existence in the dictionaries
    my  $replace        =   _search_term( $page_name, $term );

    # if we have an entry ...
    if (defined $replace) {
        my  $counter    =   _inc_counter($term);
        my  $subkey     =    $counter == 1 ? q(first) : q(replace);

        return $replace->{$subkey};
    }
    else {
        return undef;
    }
}

sub _search_term {
    my  $page_name      =   shift;
    my  $term           =   shift;

    if (defined $page_name and exists $_local_dict{$page_name}) {
        if (exists $_local_dict{$page_name}->{terms}->{$term}) {
            return $_local_dict{$page_name}->{terms}->{$term};
        }
    }

    if (exists $_dictionary{$term}) {
        return $_dictionary{$term};
    }

    return undef;
}

sub _data_dumper {
    my  $operation      =   shift;
    my  $params_ref     =   shift;

    print "sar: ${operation}\n";
    print Data::Dumper->Dump(
                [ $params_ref, \%_dictionary, \%_local_dict ],
                [ qw(params _dictionary _local_dict ) ]
                );

    return;
}

1; # Magic true value required at end of module
__END__

=head1 NAME

IkiWiki::Plugin::sar - Search And Replace terms

=head1 SYNOPSIS

In a ikiwiki source page we can write this

    [[sar]]

for enable the sar plugin in this page or

    [[sar search=debian replace="__Debian__"]]

to define a global replace for the expression I<--debian--> or 

    [[sar search=ibm first='[IBM](http://www.ibm.com)' 
        next="_IBM_"]]

to define a replace for the first match of the string I<--ibm--> and a different
replace for the rest.
    
=head1 DESCRIPTION

This plugin is useful to make global or local search and replace operations
using common or specific terms. 

The principal characteristic is that it is possible to replace the first
appearance with a text and the rest with other.

A I<search and replace> directive has the following parameters:

=over

=item search 

Define the text to search.

=item first

Define the replace text in the first match.

=item next

Define the replace text in all matches except the first

=item replace

Define the replace text in all matches.

=back

=head2 Global dictionary

In the global dictionary page we can define the global terms for the plugin:

    ## Sites and projects 

    - [[sar search="ikiwiki" first="[IkiWiki](http://ikiwiki.info)" next="_IkiWiki_"]]
    - [[sar search="debian" first="[Debian](http://debian.org)" next="_Debian_"]]
    - [[sar search="perl" first="[Perl](http://perl.org)" next="_Perl_"]]
    - [[sar search="linux" replace="GNU/Linux"]]
    - [[sar search="subversion" first="[[Subversion|software/subversion]]" 
            next="_Subversion_" ]]

    ## Persons            
    - [[sar search="joey" first="[Joey Hess](http://ikiwiki.info/users/joey]]" next="_Joey_" ]]
    - [[sar search="angel" first="[Angel](http://triptico.com)" next="Angel"]]

    ## Technical terms 

    - [[sar search="smtp" first="[[wp SMTP]]" next="`SMTP`"]]
    - [[sar search="pop3" first="[[wp POP3]]" next="'POP3'"]]

=head2 Caveats

The plugin works as a filter hook on the source markdown text; the dictionary
terms must be carefully selected, because they could interfere with URLs
and other preprocessor directives. 

=head1 FUNCTIONS

=head2 checkconfig( )

This function load the configuration and preprocess the main page for feed the
dictionary.

=head2 preprocess( ) 

This function preprocess a I<sar> directive and feed the global or private
dictionary.

=head2 filter( )

This function search the dictionary terms in a ikiwiki source page and process
it.

=head1 DIAGNOSTICS

This module don't raise any exceptions. It can work if the global page is not
found.

=head1 CONFIGURATION AND ENVIRONMENT

The plugin need the following values in the ikiwiki setup file:

=over

=item sar_pagespec

Define the pagespec expresion to select a set of pages on which working. By
default is C<*>. A good value for it could be C<link(tag/sar)>; i mean only
works on pages tagged with the I<sar> tag.

=item sar_mainpage

Define the name of the global dictionary for the plugin. By default is C<sar>.

=back

=head1 DEPENDENCIES

=over 

=item L<IkiWiki>

=item L<List::MoreUtils>

=back

=head1 INCOMPATIBILITIES

None reported.

=head1 BUGS AND LIMITATIONS

Please report any bugs or feature requests to
the author at C<victor@taquiones.net>.

Patches, corrections and criticims are welcome.

=head1 AUTHOR

"Vctor Moral"  C<< <victor@taquiones.net> >>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2007 <Victor Moral>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

